/*

cc -g sam_comp2.c -DHAVE_LIBLZMA -I/software/badger/opt/xz_utils/include -L/software/badger/opt/xz_utils/lib -lz -llzma

See end for sizes

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <zlib.h>

#define ABS(i) ((i)>0?(i):-(i))
#define COMP_MODE_NONE 0
#define COMP_MODE_ZLIB 1
#define COMP_MODE_LZMA 2

#define SIMPLE_READGROUP_MODE
//#define SIMPLE_QNAME_MODE
#define SIMPLE_CIGAR_MODE
#define SIMPLE_SEQ_MODE
#define SIMPLE_QUAL_MODE
//#define SIMPLE_AUX_MODE

static int part_uncomp[100];
static int part_comp[100];

/*
 * Dummy nul-compression functions. These do nothing except satisfy
 * the internal API requirements. We use them simply to compute the
 * base-line so we can estimate the overhead due to zlib vs overhead
 * due to lzma.
 */
static char *nul_mem_deflate(int level, char *data, size_t size, size_t *cdata_size) {
    char *out = malloc(size);
    memcpy(out, data, size);
    *cdata_size = size;
    return out;
}

static char *nul_mem_deflate_parts(int level, char *data,
				   size_t *part_size, int nparts,
				   size_t *cdata_size) {
    size_t tot_size = 0;
    int i;
    for (i = 0; i < nparts; i++)
	tot_size += part_size[i];
    return nul_mem_deflate(level, data, tot_size, cdata_size);
}

static char *nul_mem_deflate_lparts(char *data,
				    size_t *part_size, int *level, int nparts,
				    size_t *cdata_size) {
    return nul_mem_deflate_parts(*level, data, part_size, nparts, cdata_size);
}

static char *nul_mem_inflate(char *cdata, size_t csize, size_t *size) {
    char *out = malloc(csize);
    memcpy(out, cdata, csize);
    *size = csize;
    return out;
}

#ifdef HAVE_LIBLZMA
/* ------------------------------------------------------------------------ */
/*
 * Data compression routines using liblzma (xz)
 *
 * On a test set this shrunk the main db from 136157104 bytes to 114796168, but
 * caused tg_index to grow from 2m43.707s to 15m3.961s. Exporting as bfastq
 * went from 18.3s to 36.3s. So decompression suffers too, but not as bad
 * as compression times.
 *
 * For now we disable this functionality. If it's to be reenabled make sure you
 * improve the mem_inflate implementation as it's just a test hack at the
 * moment.
 */
#include <lzma.h>

/* Fast mode, but not too bad on compression still */
//#define LZMA_LEVEL 3

static char *lzma_mem_deflate(int level, char *data, size_t size, size_t *cdata_size) {
    char *out;
    size_t out_size = lzma_stream_buffer_bound(size);
    *cdata_size = 0;

    out = malloc(out_size);

    /* Single call compression */
    if (LZMA_OK != lzma_easy_buffer_encode(level, LZMA_CHECK_CRC32, NULL,
					   (uint8_t *)data, size,
					   (uint8_t *)out, cdata_size,
					   out_size))
    	return NULL;

    return out;
}

static char *lzma_mem_deflate_parts(int level, char *data,
				    size_t *part_size, int nparts,
				    size_t *cdata_size) {
    size_t tot_size = 0;
    int i;
#if 0
    for (i = 0; i < nparts; i++)
	tot_size += part_size[i];
    return lzma_mem_deflate(level, data, tot_size, cdata_size);

#else
    char  *lzo[100], *out;
    size_t lzs[100];

    /* Compress parts */
    for (i = 0; i < nparts; i++) {
	lzo[i] = lzma_mem_deflate(level, data, part_size[i], &lzs[i]);
	//fprintf(stderr, "part %2d, size %7ld, comp_size %7ld, fract %f\n",
	//	i, part_size[i], lzs[i], (100.0 * lzs[i]) /
	//	part_size[i]);
	part_uncomp[i] += part_size[i];
	part_comp[i] += lzs[i];
	data += part_size[i];
	tot_size += lzs[i];
    }

    /* Stitch together */
    data = out = malloc(tot_size);
    //fprintf(stderr, "=> total size %ld\n", tot_size);
    for (i = 0; i < nparts; i++) {
	memcpy(data, lzo[i], lzs[i]);
	free(lzo[i]);
	data += lzs[i];
    }

    *cdata_size = tot_size;
    return out;
#endif
}

static char *lzma_mem_deflate_lparts(char *data,
				     size_t *part_size, int *level, int nparts,
				     size_t *cdata_size) {
    return lzma_mem_deflate_parts(*level, data, part_size, nparts, cdata_size);
}

static char *lzma_mem_inflate(char *cdata, size_t csize, size_t *size) {
    lzma_stream strm = LZMA_STREAM_INIT;
    size_t out_size = 0, out_pos = 0;
    char *out = NULL;
    int r;

    /* Initiate the decoder */
    if (LZMA_OK != lzma_stream_decoder(&strm, 50000000, 0))
	return NULL;

    /* Decode loop */
    strm.avail_in = csize;
    strm.next_in = (uint8_t *)cdata;

    for (;strm.avail_in;) {
	if (strm.avail_in > out_size - out_pos) {
	    out_size += strm.avail_in * 4 + 32768;
	    out = realloc(out, out_size);
	}
	strm.avail_out = out_size - out_pos;
	strm.next_out = (uint8_t *)&out[out_pos];

	r = lzma_code(&strm, LZMA_RUN);
	if (LZMA_OK != r && LZMA_STREAM_END != r) {
	    fprintf(stderr, "r=%d\n", r);
	    fprintf(stderr, "mem=%ld\n", lzma_memusage(&strm));
	    return NULL;
	}

	out_pos = strm.total_out;

	if (r == LZMA_STREAM_END)
	    break;
    }

    /* finish up any unflushed data; necessary? */
    r = lzma_code(&strm, LZMA_FINISH);
    if (r != LZMA_OK && r != LZMA_STREAM_END) {
	fprintf(stderr, "r=%d\n", r);
	return NULL;
    }

    out = realloc(out, strm.total_out);
    *size = strm.total_out;

    lzma_end(&strm);

    return out;
}
#endif

/* ------------------------------------------------------------------------ */
/*
 * Data compression routines using zlib.
 */
#include <zlib.h>
static char *zlib_mem_deflate(int level, char *data, size_t size, size_t *cdata_size) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err;

    cdata = malloc(cdata_alloc = size*1.05+10);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.avail_in = size;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, level, Z_DEFLATED, 15, 9, Z_DEFAULT_STRATEGY);

    /* Encode to 'cdata' array */
    for (;s.avail_in;) {
	s.next_out = &cdata[cdata_pos];
	s.avail_out = cdata_alloc - cdata_pos;
	if (cdata_alloc - cdata_pos <= 0) {
	    fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
	    return NULL;
	}
	err = deflate(&s, Z_NO_FLUSH);
	cdata_pos = cdata_alloc - s.avail_out;
	if (err != Z_OK) {
	    fprintf(stderr, "zlib deflate error: %s\n", s.msg);
	    break;
	}
    }
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    *cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    return (char *)cdata;
}

static char *zlib_mem_deflate_parts(int level, char *data,
				    size_t *part_size, int nparts,
				    size_t *cdata_size) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err, i;
    size_t size = 0;

    int strat[14] = {Z_DEFAULT_STRATEGY,
		     Z_DEFAULT_STRATEGY,
		     Z_FILTERED,
		     Z_FILTERED,
		     Z_DEFAULT_STRATEGY,
		     Z_DEFAULT_STRATEGY,
		     Z_FILTERED,
		     Z_FILTERED,
		     Z_FILTERED,
		     Z_FILTERED,
		     Z_FILTERED,
		     Z_FILTERED,
		     Z_FILTERED, //Z_RLE or Z_HUFFMAN_ONLY also work ok,
		     Z_FILTERED};

    for (i = 0; i < nparts; i++)
	size += part_size[i];

    cdata = malloc(cdata_alloc = size*1.05+256*nparts);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, level, Z_DEFLATED, 15,
		       9/*mem level */, Z_FILTERED);

    /* Encode to 'cdata' array */
    for (i = 0; i < nparts; i++) {
	deflateParams(&s, level, strat[i]);

	s.avail_in = part_size[i];
	for (;s.avail_in;) {
	    s.next_out = &cdata[cdata_pos];
	    s.avail_out = cdata_alloc - cdata_pos;
	    if (cdata_alloc - cdata_pos <= 0) {
		fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
		return NULL;
	    }
	    err = deflate(&s, Z_SYNC_FLUSH); // also try Z_FULL_FLUSH
	    if (1) {
		static int last;
		if (i == 0) last = 0;
		int sz = s.total_out - last;
		last = s.total_out;
		part_uncomp[i] += part_size[i];
		part_comp[i] += sz;
		/*
		fprintf(stderr,
			"part %2d, size %7ld, comp_size %7ld, fract %f\n",
			i, part_size[i], sz, (100.0 * sz) / part_size[i]);
		*/
	    }
	    cdata_pos = cdata_alloc - s.avail_out;
	    if (err != Z_OK) {
		fprintf(stderr, "zlib deflate error: %s\n", s.msg);
		break;
	    }
	}
    }
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    *cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    return (char *)cdata;
}

static char *zlib_mem_deflate_lparts(char *data,
				     size_t *part_size, int *level, int nparts,
				     size_t *cdata_size) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err, i;
    size_t size = 0;

    for (i = 0; i < nparts; i++)
	size += part_size[i];

    cdata = malloc(cdata_alloc = size*1.05+256*nparts);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15,
		       9, Z_FILTERED);

    /* Encode to 'cdata' array */
    for (i = 0; i < nparts; i++) {
	s.avail_in = part_size[i];
	for (;s.avail_in;) {
	    s.next_out = &cdata[cdata_pos];
	    s.avail_out = cdata_alloc - cdata_pos;
	    if (cdata_alloc - cdata_pos <= 0) {
		fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
		return NULL;
	    }
	    deflateParams(&s, level[i], Z_DEFAULT_STRATEGY);
	    err = deflate(&s, Z_SYNC_FLUSH); // also try Z_FULL_FLUSH
	    cdata_pos = cdata_alloc - s.avail_out;
	    if (err != Z_OK) {
		fprintf(stderr, "zlib deflate error: %s\n", s.msg);
		break;
	    }
	}
    }
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    *cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    return (char *)cdata;
}

static char *zlib_mem_inflate(char *cdata, size_t csize, size_t *size) {
    z_stream s;
    unsigned char *data = NULL; /* Uncompressed output */
    int data_alloc = 0;
    int err;

    /* Starting point at uncompressed size, 4x compressed */
    data = malloc(data_alloc = csize*4+10);

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)cdata;
    s.avail_in = csize;
    s.total_in = 0;
    s.next_out  = data;
    s.avail_out = data_alloc;
    s.total_out = 0;

    err = inflateInit(&s);

    /* Decode to 'data' array */
    for (;s.avail_in;) {
	s.next_out = &data[s.total_out];
	err = inflate(&s, Z_NO_FLUSH);
	if (err == Z_STREAM_END)
	    break;

	if (err != Z_OK) {
	    fprintf(stderr, "zlib inflate error: %s\n", s.msg);
	    break;
	}

	/* More to come, so realloc */
	data = realloc(data, data_alloc += s.avail_in*4 + 10);
	s.avail_out += s.avail_in*4+10;
    }
    inflateEnd(&s);

    *size = s.total_out;
    return (char *)data;
}

static char *mem_deflate(int mode, int level,
			 char *data, size_t size, size_t *cdata_size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_deflate (level, data, size, cdata_size);
    case COMP_MODE_ZLIB:
	return zlib_mem_deflate(level, data, size, cdata_size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_deflate(level, data, size, cdata_size);
#endif	
    }
    
    return NULL;
}

static char *mem_deflate_parts(int mode, int level, char *data,
			       size_t *part_size, int nparts,
			       size_t *cdata_size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_deflate_parts (level, data, part_size, nparts, cdata_size);
    case COMP_MODE_ZLIB:
	return zlib_mem_deflate_parts(level, data, part_size, nparts, cdata_size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_deflate_parts(level, data, part_size, nparts, cdata_size);
#endif	
    }
    
    return NULL;
}

static char *mem_deflate_lparts(int mode, char *data,
				size_t *part_size, int *level, int nparts,
				size_t *cdata_size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_deflate_lparts (data, part_size, level, nparts, cdata_size);
    case COMP_MODE_ZLIB:
	return zlib_mem_deflate_lparts(data, part_size, level, nparts, cdata_size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_deflate_lparts(data, part_size, level, nparts, cdata_size);
#endif	
    }
    
    return NULL;
}

static char *mem_inflate(int mode,
			 char *cdata, size_t csize, size_t *size) {
    switch (mode) {
    case COMP_MODE_NONE:
	return nul_mem_inflate (cdata, csize, size);
    case COMP_MODE_ZLIB:
	return zlib_mem_inflate(cdata, csize, size);
#ifdef HAVE_LIBLZMA
    case COMP_MODE_LZMA:
	return lzma_mem_inflate(cdata, csize, size);
#endif	
    }
    
    return NULL;
}

/*
 * Converts an unsigned integer value into a 7-bit encoded format.
 * We store the bottom 7 bits of value with either 0 or 1 for the top-bit
 * depending on whether any bits are left. We keep repeating this until
 * all significant bits of value have been used.
 *
 * Ie 15551 = hex 3cbf = 0011 1100 1011 1111 becomes:
 *
 *     111 1001   011 1111 (0x3cbf input)
 *    0111 1001  1011 1111 (0x79bf output)
 *
 * Takes an unsigned 32-bit integer and stores in out.
 * Returns the number of bytes written to 'out'
 */
int int2u7(uint32_t in, unsigned char *out) {
    int n = 0;

    out[0] = in & 0x7f;
    while (in >= 128) {
	out[n++] |= 128;
	in >>= 7;
	out[n] = in & 0x7f;
    }

    return n+1;
}

/*
 * Takes a 7-bit encoded value in 'u7' and stores in a
 * 32-bit unsigned int pointed to by 'out'.
 *
 * Returns the number of bytes read from u7.
 */
int u72int(unsigned char *u7, uint32_t *out) {
    uint32_t ret = 0;
    int b = 0;

    ret |= *u7 & 0x7f;
    while (*u7 & 0x80) {
	u7++;
	b += 7;
	ret |= (*u7 & 0x7f) << b;
    }

    *out = ret;
    return b/7+1;
}

/*
 * Converts an signed value into a 7-bit encoded format.
 * We store the bottom 7 bits of value with either 0 or 1 for the top-bit
 * depending on whether any bits are left. We keep repeating this until
 * all significant bits of value have been used.
 *
 * Ie 15551 = hex 3cbf = 0011 1100 1011 1111 becomes:
 *
 *     111 1001   011 1111 (0x3cbf input)
 *    0111 1001  1011 1111 (0x79bf output)
 *
 * Takes an unsigned 32-bit integer and stores in out.
 * Returns the number of bytes written to 'out'
 */
int int2s7(int32_t in, unsigned char *out) {
    int n = 0;
    uint32_t u = (ABS(in) << 1) | (in < 0);

    out[0] = u & 0x7f;
    while (u >= 128) {
	out[n++] |= 128;
	u >>= 7;
	out[n] = u & 0x7f;
    }

    return n+1;
}

/*
 * Takes a 7-bit encoded value in 'u7' and stores in a
 * 32-bit signed int pointed to by 'out'.
 *
 * Returns the number of bytes read from u7.
 */
int s72int(unsigned char *u7, int32_t *out) {
    uint32_t ret = 0;
    int b = 0;

    ret |= *u7 & 0x7f;
    while (*u7 & 0x80) {
	u7++;
	b += 7;
	ret |= (*u7 & 0x7f) << b;
    }

    *out = (ret & 1) ? -(ret >> 1) : (ret >> 1);
    return b/7+1;
}

/* Dynamic growing string */
typedef struct {
    char *str;
    size_t alloc;
    size_t sz;
} dstr;

dstr *dstr_create(void) {
    dstr *ds = malloc(sizeof(*ds));
    if (!ds)
	return NULL;

    ds->str = NULL;
    ds->alloc = 0;
    ds->sz = 0;

    return ds;
}

char *dstr_str(dstr *ds) {
    return ds->str ? ds->str : "";
}

void dstr_empty(dstr *ds) {
    ds->sz = 0;
    if (ds->str)
	*ds->str = 0;
}

void dstr_free(dstr *ds) {
    if (ds) {
	if (ds->str)
	    free(ds->str);
	free(ds);
    }
}

int dstr_append_char(dstr *ds, char c) {
    char *cp;

    if (ds->sz + 2 > ds->alloc) {
	ds->alloc += 8192;
	cp = realloc(ds->str, ds->alloc);
	if (!cp)
	    return -1;
	ds->str = cp;
    }

    cp = &ds->str[ds->sz++];
    *cp++ = c;
    *cp++ = 0;

    return 0;
}

int dstr_append_strn(dstr *ds, char *str, size_t len) {
    char *cp;

    if (ds->sz + len + 1 > ds->alloc) {
	while (ds->sz + len + 1 > ds->alloc) {
	    ds->alloc += 8192;
	}
	cp = realloc(ds->str, ds->alloc);
	if (!cp)
	    return -1;
	ds->str = cp;
    }
    cp = &ds->str[ds->sz];
    memcpy(cp, str, len);
    cp[len] = 0;

    ds->sz += len;

    return 0;
}

int dstr_append_int(dstr *ds, int i) {
    char *cp, *cpo;
    int j, k = 0;

    if (ds->sz + 20 > ds->alloc) {
	ds->alloc += 8192;
	cp = realloc(ds->str, ds->alloc);
	if (!cp)
	    return -1;
	ds->str = cp;
    }

    cpo = cp = &ds->str[ds->sz];

    if (i < 0) {
	*cp++ = '-';
	i = -i;
    } else if (i == 0) {
	*cp++ = '0';
	ds->sz += cp - cpo;
	*cp++ = 0;
	return 0;
    }

    if (i < 1000)
	goto b1;
    if (i < 100000)
	goto b2;
    if (i < 100000000)
	goto b3;

    j = i / 1000000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000000000;

    j = i / 100000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 100000000;
    
 b3:
    j = i / 10000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 10000000;
    
    j = i / 1000000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000000;
    
    j = i / 100000;
    if (j || k) *cp++ = j + '0', k=1, i %= 100000;
    
 b2:
    j = i / 10000;
    if (j || k) *cp++ = j + '0', k=1, i %= 10000;

    j = i / 1000;
    if (j || k) *cp++ = j + '0', k=1, i %= 1000;

 b1:
    j = i / 100;
    if (j || k) *cp++ = j + '0', k=1, i %= 100;

    j = i / 10;
    if (j || k) *cp++ = j + '0', k=1, i %= 10;

    if (i || k) *cp++ = i + '0';

    ds->sz += cp - cpo;
    *cp++ = 0;

    return 0;
}

typedef struct {
    int type;
    char format;
    int  len;
    char *val;
} aux_t;

typedef struct sam_entry {
    int flag, pos, mapq, mpos, isize;
    char *qname, *rname, *cigar, *mrnm, *seq, *qual;
    int qname_len, rname_len, cigar_len, mrnm_len, seq_len;
    char *readgroup;
    aux_t a[50];
    int a_cnt;
    int delayed;
} sam_entry;

/* Fetches the next string from *cp */
char *get_next_str(char **cp, int *lenp) {
    char *c = *cp;
    char *str = c;
    int len = 0;

    while (*c && *c != '\t') {
	c++;
	len++;
    }
    if (*c == '\t')
	*c++ = 0;

    *cp = c;
    if (lenp)
	*lenp = len;

    return str;
}

/* Fetches the next integer from *cp */
int get_next_int(char **cp) {
    int i = 0, neg = 0;
    char *c = *cp;

    if (*c == '-') {
	neg = 1;
	c++;
    } else if (*c == '+') {
	c++;
    }

    while (*c >= '0' && *c <= '9')
	i = i*10 + *c++ - '0';
    if (*c != 0 && *c != '\t') {
	fprintf(stderr, "Malformed integer '%s'\n", *cp);
	exit(1);
    }

    if (*c == '\t')
	c++;

    *cp = c;
    return neg ? -i : i;
}

#define AUX_TYPE(a,b) (((a)<<8) | (b))
#define AUX_TYPE_RG AUX_TYPE('R','G')
#define AUX_TYPE_OQ AUX_TYPE('O','Q')

int parse_aux(char *cp, sam_entry *sam) {
    sam->a_cnt = 0;
    sam->readgroup = NULL;

    while (cp[0] != '\0' && cp[1] != '\0') {
	/* Parse a single aux entry */
	int type = (cp[0]<<8) + cp[1];
	char *val;
	int val_len;
	int vtype;

	cp += 2;
	if (*cp++ != ':')
	    return -1;

	vtype = *cp++;

	if (*cp++ != ':')
	    return -1;
	val = cp;
	while (*cp != '\t' && *cp != '\0' && *cp != '\n')
	    cp++;
	val_len = cp-val;

	while (*cp == '\t' || *cp == '\n')
	    *cp++ = 0;

	/* Store */
	/*
	printf("type=%c%c, vtype=%c, val='%.*s'\n",
	       (type >> 8), type & 0xff,
	       vtype,
	       val_len, val);
	*/
	sam->a[sam->a_cnt].type   = type;
	sam->a[sam->a_cnt].format = vtype;
	sam->a[sam->a_cnt].val    = val;
	sam->a[sam->a_cnt].len    = val_len;
	sam->a_cnt++;

	if (type == AUX_TYPE_RG)
	    sam->readgroup     = val;
    }

    return 0;
}

typedef struct {
    sam_entry *sam;
    int idx;
} sam_order;

int sam_order_cmp(const void *p1, const void *p2) {
    const sam_order *s1 = (const sam_order *)p1;
    const sam_order *s2 = (const sam_order *)p2;

    if (!s1->sam->readgroup && !s2->sam->readgroup) {
	return s1->sam - s2->sam;
    } else if (!s1->sam->readgroup) {
	return -1;
    } else if (!s2->sam->readgroup) {
	return +1;
    } else {
	int d = strcmp(s1->sam->readgroup, s2->sam->readgroup);
	if (d == 0) {
	    return s1->sam - s2->sam;
	} else {
	    return d;
	}
    }
}

typedef struct aux_hist {
    char aux[1024];
    int len;
    struct aux_hist *next;
} aux_hist_t;

#define NHIST 10
#define LINE_BLK 8*1024

void comp(FILE *fpi, FILE *fpo, int count, int mode, int level) {
    int h, i, j, sz = 0;
    char **lines, *cp, *cp2, *data, *last_str;
    int *pos;
    char line[65536];
    sam_entry *sam;
    sam_order *ord;
    int last;
    char lookup[256];
    int nl = 0;

    char aux_enc1[256], aux_enc2[256], aux_enc3[256];
    char aux_enc[256], *aux_p;

    int aux_len1, aux_len2, aux_len3, aux_len;
    aux_hist_t aux_hist[NHIST], *aux_hist_p, *aux_hist_f, *aux_pr, *aux_cp,
	*aux_pr2;

    /* Create lookup table to convert ACGT? to 01234*/
    memset(lookup, 5, 256);
    lookup['A'] = 0;
    lookup['C'] = 1;
    lookup['G'] = 2;
    lookup['T'] = 3;
    lookup['N'] = 4;

    /* Zam header */
    {
	char hdr[16];
	memcpy(hdr, ".zam1.00", 8);
	hdr[8] = mode;
	memset(&hdr[9], 0, 7);
	fwrite(hdr, 1, 16, fpo);
	//write(1, hdr, 16);
    }

    /* Sam header */
    {
	char *hdr = malloc(1000000); /* fixme: check actual size */
	char *cp = hdr, *gzout;
	int c;
	uint32_t u32;
	size_t ssz;

	while ((c = fgetc(fpi)) == '@') {
	    line[0] = '@';
	    fgets(line+1, 65535, fpi);
	    strcpy(cp, line);
	    cp += strlen(cp);
	}
	ungetc(c, fpi);

	switch(mode) {
	case COMP_MODE_LZMA:
	case COMP_MODE_ZLIB:
	    gzout = mem_deflate(mode, level, hdr, cp-hdr, &ssz);
	    free(hdr);
	    break;

	default:
	    gzout = hdr;
	    ssz = cp-hdr;
	}

	/* FIXME: endian specific */
	u32 = ssz;
	fwrite(&u32, 4, 1, fpo);
	fwrite(gzout, 1, ssz, fpo);
	//write(1, &u32, 4);
	//write(1, gzout, ssz);
	free(gzout);
    }

    /* Gather data */
    lines = (char **)malloc(count * sizeof(char *));
    pos   = (int *)malloc(count * sizeof(int));
    sam = (sam_entry *)malloc(count * sizeof(*sam));
    ord = (sam_order *)malloc(count * sizeof(*ord));

    for (;;) {
	size_t blk_size[20];
	size_t ssz;
	char *gzout;
	size_t nlen = 0;
	uint32_t u32;
	int osz;

	nl = 0;
	lines[0] = malloc(LINE_BLK);
	pos[0] = 0;

	for (i = 0; i < count; i++) {
	    size_t line_len;

	    if (fgets(line, 65536, fpi) == NULL)
		break;

	    line_len = strlen(line);
	    if (line_len + pos[nl] >= LINE_BLK) {
		lines[++nl] = malloc(LINE_BLK);
		pos[nl] = 0;
	    }

	    cp = lines[nl] + pos[nl];
	    strcpy(cp, line);
	    //memcpy(cp, line, line_len); cp[line_len] = 0;
	    pos[nl] += line_len;
	    sz += line_len;

	    /* Parse the sam line */
	    sam[i].qname = get_next_str(&cp, &sam[i].qname_len);
	    nlen += sam[i].qname_len;
	    sam[i].flag  = get_next_int(&cp);
	    sam[i].rname = get_next_str(&cp, &sam[i].rname_len);
	    sam[i].pos   = get_next_int(&cp);
	    sam[i].mapq  = get_next_int(&cp);
	    sam[i].cigar = get_next_str(&cp, &sam[i].cigar_len);
	    sam[i].mrnm  = get_next_str(&cp, &sam[i].mrnm_len);
	    sam[i].mpos  = get_next_int(&cp);
	    sam[i].isize = get_next_int(&cp);
	    sam[i].seq   = get_next_str(&cp, &sam[i].seq_len);
	    sam[i].qual  = get_next_str(&cp, NULL);

	    /* Parse the aux line. FIXME: todo */
	    parse_aux(cp, &sam[i]);

	    /*
	      printf("qname = '%s'\n", sam[i].qname);
	      printf("flag  = '%d'\n", sam[i].flag);
	      printf("rname = '%s'\n", sam[i].rname);
	      printf("pos   = '%d'\n", sam[i].pos);
	      printf("mapq  = '%d'\n", sam[i].mapq);
	      printf("cigar = '%s'\n", sam[i].cigar);
	      printf("mrnm  = '%s'\n", sam[i].mrnm);
	      printf("mpos  = '%d'\n", sam[i].mpos);
	      printf("isize = '%d'\n", sam[i].isize);
	      printf("seq   = '%s'\n", sam[i].seq);
	      printf("qual  = '%s'\n", sam[i].qual);
	      printf("aux   = '%s'\n", sam[i].aux);
	    */
	    ord[i].idx = i;
	    ord[i].sam = &sam[i];
	}
	count = i;
	
	if (count == 0) {
	    for (i = 0; i <= nl; i++)
		free(lines[i]);
	    break;
	}

	/* Dump out in our new order */
	data = cp = cp2 = malloc(sz);

	/*-- Read group */
#ifdef SIMPLE_READGROUP_MODE
	for (i = 0; i < count; i++) {
	    if (sam[i].readgroup) {
		size_t l = strlen(sam[i].readgroup);
		memcpy(cp, sam[i].readgroup, l);
		cp += l;
		*cp++ = 0;
	    } else {
		*cp++ = 0;
	    }
	}
#else
	{
	    char *last = NULL;
	    size_t j, last_len = 0;

	    for (i = 0; i < count; i++) {
		size_t l;

		if (!sam[i].readgroup) {
		    *cp++ = -1;
		    continue;
		}

		l = strlen(sam[i].readgroup);
		for (j = 0; j < last_len && j < l; j++) {
		    if (sam[i].readgroup[j] != last[j])
			break;
		}
		*cp++ = j;

		memcpy(cp, sam[i].readgroup+j, l-j);
		cp += l-j;
		*cp++ = 0;

		last = sam[i].readgroup;
		last_len = l;
	    }
	}
#endif

	blk_size[0] = cp-cp2; cp2 = cp;

	/* Sort by read group too so we can optimise later on */
	qsort(ord, count, sizeof(ord[0]), sam_order_cmp);

	/*
	for (i = 0; i < count; i++) {
	    printf("%d %d %s\n",
		   i, ord[i].idx, ord[i].sam->readgroup);
	}
	*/

	/*-- Flag */
	for (i = 0; i < count; i++) {
	    cp += int2u7(sam[i].flag, (unsigned char *)cp);
	}
	blk_size[1] = cp-cp2; cp2 = cp;

	/*-- Pos */
	last = 1;
	for (i = 0; i < count; i++) {
	    cp += int2u7(sam[i].pos - last, (unsigned char *)cp);
	    last = sam[i].pos;
	}
	blk_size[2] = cp-cp2; cp2 = cp;

	/*-- MapQ */
	for (i = 0; i < count; i++) {
	    //if ((sam[i].flag & 0x04) == 4) continue;
	    cp += int2u7(sam[i].mapq, (unsigned char *)cp);
	}
	blk_size[3] = cp-cp2; cp2 = cp;

	/*-- Mate Pos */
	last = 1;
	for (i = 0; i < count; i++) {
	    if (sam[i].mpos == 0) {
		*cp++ = 0;
	    } else {
		int v = (sam[i].mpos - sam[i].pos)*2+1;
		cp += int2s7(v, (unsigned char *)cp);
	    }
	}
	blk_size[4] = cp-cp2; cp2 = cp;

	/*-- Insert size */
	for (i = 0; i < count; i++) {
	    //if ((sam[i].flag & 0x04) == 4) continue;
	    cp += int2s7(sam[i].isize, (unsigned char *)cp);
	}
	blk_size[5] = cp-cp2; cp2 = cp;

	/*-- Query name */
#ifdef SIMPLE_QNAME_MODE
	for (i = 0; i < count; i++) {
	    memcpy(cp, ord[i].sam->qname, ord[i].sam->qname_len);
	    cp += ord[i].sam->qname_len;
	    *cp++ = 0;
	}
#else
	{
	    char *last = NULL;
	    int j, last_len = 0;

	    for (i = 0; i < count; i++) {
		for (j = 0; j < last_len && j < ord[i].sam->qname_len; j++) {
		    if (ord[i].sam->qname[j] != last[j])
			break;
		}
		*cp++ = j;

		memcpy(cp, ord[i].sam->qname+j, ord[i].sam->qname_len-j);
		cp += ord[i].sam->qname_len-j;
		*cp++ = 0;

		last = ord[i].sam->qname;
		last_len = ord[i].sam->qname_len;
	    }
	}
#endif
	blk_size[6] = cp-cp2; cp2 = cp;

	/*-- Reference name */
	last_str = "";
	for (i = 0; i < count; i++) {
	    if (strcmp(sam[i].rname, last_str) == 0) {
		cp += int2u7(0, (unsigned char *)cp);
	    } else {
		cp += int2u7(sam[i].rname_len, (unsigned char *)cp);
		memcpy(cp, sam[i].rname, sam[i].rname_len);
		cp += sam[i].rname_len;
	    }
	    last_str = sam[i].rname;
	}
	blk_size[7] = cp-cp2; cp2 = cp;

	/*-- Cigar */
#ifdef SIMPLE_CIGAR_MODE
	for (i = 0; i < count; i++) {
	    //if ((sam[i].flag & 0x04) == 4) continue;

#if 0
	    /* Encode * and <len>M as special cigar cases */
	    if (ord[i].sam->cigar[0] == '*') {
		*cp++ = 0;
		continue;
	    } else {
		char b[100];
		sprintf(b, "%dM", ord[i].sam->seq_len);
		if (strncmp(b, ord[i].sam->cigar, ord[i].sam->cigar_len) == 0) {
		    *cp++ = 1;
		    continue;
		}
	    }
#endif

	    memcpy(cp, ord[i].sam->cigar, ord[i].sam->cigar_len);
	    cp += ord[i].sam->cigar_len;
	    *cp++ = 0;
	}
#else
	/* Currently very minimal benfit - rewrite */
	for (i = 0; i < count; i++) {
	    char *ncp1;

	    cp += int2u7(sam[i].cigar_len, (unsigned char *)cp);
	    ncp1 = sam[i].cigar;
	    for (j = 0; j < sam[i].cigar_len; j++) {
		int k = 0;
		while (ncp1[j] >= '0' && ncp1[j] <= '9')
		    k = k*10 + ncp1[j++] - '0';
		cp += int2u7(k, (unsigned char *)cp);
		*cp++ = ncp1[j];
	    }
	}
#endif
	blk_size[8] = cp-cp2; cp2 = cp;

	/*-- Mate reference name */
	for (i = 0; i < count; i++) {
	    memcpy(cp, sam[i].mrnm, sam[i].mrnm_len);
	    cp += sam[i].mrnm_len;
	    *cp++ = 0;
	}
	blk_size[9] = cp-cp2; cp2 = cp;

	/*-- Seq. length */
	for (i = 0; i < count; i++) {
	    cp += int2u7(ord[i].sam->seq_len, (unsigned char *)cp);
	}
	blk_size[10] = cp-cp2; cp2 = cp;

	/*-- Sequence */
#ifdef SIMPLE_SEQ_MODE
	for (i = 0; i < count; i++) {
	    if ((sam[i].flag & 0x04) == 0) {
		memcpy(cp, sam[i].seq, sam[i].seq_len); cp += sam[i].seq_len;
	    }
	}

	for (i = 0; i < count; i++) {
	    if ((sam[i].flag & 0x04) != 0) {
		memcpy(cp, sam[i].seq, sam[i].seq_len); cp += sam[i].seq_len;
	    }
	}
#else	/*
	 * Encode sequence in triplets. This works best is we pad out data
	 * to align based on pos%3, so that a sequence shifted by 1 byte
	 * still mostly matches the previous encoded string.
	 *
	 * NB: doesn't seem to shrink data at all.
	 * Maybe use a better delta coding system instead?
	 */
	for (i = 0; i < count; i++) {
	    char seq[1024], *s;
	    int l;

	    switch (sam[i].pos % 3) {
	    case 0:
		s = seq;
		memcpy(s, sam[i].seq, sam[i].seq_len);
		l = sam[i].seq_len;
		break;

	    case 1:
		s = seq+1;
		memcpy(s, sam[i].seq, sam[i].seq_len);
		l = sam[i].seq_len+3;
		seq[0] = '?';
		seq[l-1] = '?';
		seq[l-2] = '?';
		break;

	    case 2:
		s = seq+2;
		memcpy(s, sam[i].seq, sam[i].seq_len);
		l = sam[i].seq_len+3;
		seq[0] = '?';
		seq[1] = '?';
		seq[l-1] = '?';
		break;
	    }

	    for (j = 0; j < l; j+=3) {
		*cp++ = lookup[seq[j]]*36 + lookup[seq[j+1]]*6 + lookup[seq[j+2]];
	    }
	}

	/*
	for (i = 0; i < count; i++) {
	    char *s = sam[i].seq, *start = cp;
	    switch (sam[i].pos % 3) {
	    case 0:
		j = 0;
		break;
	    case 1:
		*cp++ = lookup['?']*25 + lookup[s[0]]*5 + lookup[s[1]];
		j = 2;
		break;
	    case 2:
		*cp++ = lookup['?']*25 + lookup['?']*5 + lookup[s[0]];
		j = 1;
		break;
	    }
	    for (; j < sam[i].seq_len; j+=3) {
		*cp++ = lookup[s[j]]*25 +
		    lookup[j+1 < sam[i].seq_len ? s[j+1] : '?']*5 +
		    lookup[j+2 < sam[i].seq_len ? s[j+2] : '?'];
	    }
	}
	*/
#endif
	//write(2, cp2, cp-cp2);
	blk_size[11] = cp-cp2; cp2 = cp;

	/*-- Quality */
#ifdef SIMPLE_QUAL_MODE
	/* Quality as is */
	for (i = 0; i < count; i++) {
	    memcpy(cp, ord[i].sam->qual, ord[i].sam->seq_len);
	    cp += ord[i].sam->seq_len;
	}
#else
	/* Reverse complement if needed. NB: this doesn't always help */
	for (i = 0; i < count; i++) {
	    if ((ord[i].sam->flag & 0x10) == 0) {
		memcpy(cp, ord[i].sam->qual, ord[i].sam->seq_len);
		cp += ord[i].sam->seq_len;
	    } else {
		char *q = ord[i].sam->qual;
		for (j = ord[i].sam->seq_len-1; j >= 0; j--)
		    *cp++ = q[j];
	    }
	}
#endif
	//write(2, cp2, cp-cp2);
	blk_size[12] = cp-cp2; cp2 = cp;

	/*-- Aux */
#ifdef SIMPLE_AUX_MODE
 {
	/* Mixes aux formats and values together */
	for (i = 0; i < count; i++) {
	    cp += int2u7(sam[i].a_cnt, (unsigned char *)cp);
	    for (j = 0; j < sam[i].a_cnt; j++) {
		*cp++ = sam[i].a[j].type >> 8;
		*cp++ = sam[i].a[j].type & 0xff;

		if (sam[i].a[j].type == AUX_TYPE_RG)
		    continue;

		switch (sam[i].a[j].format) {
		case 'i': {
		    int v = atoi(sam[i].a[j].val);
		    if (v == 0) {
			*cp++ = '0';
		    } else if (v > 0) {
			*cp++ = 'i';
			cp += int2u7(v, (unsigned char *)cp);
		    } else {
			*cp++ = 'I';
			cp += int2u7(-v, (unsigned char *)cp);
		    }
		    break;
		}

		case 'Z':
		    *cp++ = sam[i].a[j].format;
		    memcpy(cp, sam[i].a[j].val, sam[i].a[j].len);
		    cp += sam[i].a[j].len;
		    *cp++ = 0;
		    break;

		case 'A':
		    *cp++ = sam[i].a[j].format;
		    *cp++ = *sam[i].a[j].val;
		    break;

		default:
		    fprintf(stderr, "Unknown aux format %c\n",
			    sam[i].a[j].format);
		    break;
		}
	    }
	}
#else

   {
	/*
	 * Track recent aux formats and output symbols to represent
	 * "2nd most recent format string" instead of the actual formats.
	 * Currently we track up to 3 formats, but we may need to implement
	 * more.
	 * FIXME: Use pointers instead of memcpy too.
	 */
	for (h = 0; h < NHIST; h++) {
	    aux_hist[h].len = 0;
	    aux_hist[h].next = &aux_hist[h+1];
	}
	aux_hist[NHIST-1].next = NULL;
	aux_hist_p = NULL;         /* used list */
	aux_hist_f = &aux_hist[0]; /* free list */

	aux_enc1[0] = 0; aux_len1 = 0;
	aux_enc2[0] = 0; aux_len2 = 0;
	aux_enc3[0] = 0; aux_len3 = 0;

	for (i = 0; i < count; i++) {
	    //sam_entry *samp = &sam[i];
	    sam_entry *samp = ord[i].sam;

	    aux_p = aux_enc;
	    //aux_p += int2u7(sam[i].a_cnt, aux_p);
	    for (j = 0; j < samp->a_cnt; j++) {
		*aux_p++ = samp->a[j].type >> 8;
		*aux_p++ = samp->a[j].type & 0xff;

		if (samp->a[j].type == AUX_TYPE_RG)
		    continue;

		switch (samp->a[j].format) {
		case 'i': {
		    int v = atoi(samp->a[j].val);
		    if (v == 0) {
			*aux_p++ = '0';
		    } else if (v >= 0) {
			*aux_p++ = 'i';
		    } else {
			*aux_p++ = 'I';
		    }
		    break;
		}

		case 'A':
		case 'Z':
		    *aux_p++ = samp->a[j].format;
		    break;

		default:
		    fprintf(stderr, "Unknown aux format %c\n",
			    samp->a[j].format);
		    break;
		}
	    }

	    aux_len = aux_p - aux_enc;

	    /*
	    fprintf(stderr, "aux   =%.*s\n", aux_len, aux_enc);
	    fprintf(stderr, "  aux1=%.*s\n", aux_len1, aux_enc1);
	    fprintf(stderr, "  aux2=%.*s\n", aux_len2, aux_enc2);
	    fprintf(stderr, "  aux3=%.*s\n", aux_len3, aux_enc3);
	    */

	    for (aux_pr2 = NULL, aux_pr = NULL, aux_cp = aux_hist_p, h=1;
		 aux_cp;
		 aux_pr2 = aux_pr, aux_pr = aux_cp, aux_cp = aux_cp->next, h++) {
		if (aux_len == aux_cp->len &&
		    memcmp(aux_enc, aux_cp->aux, aux_len) == 0)
		    break;
	    }

	    //fprintf(stderr, "h=%d (%.*s)\n", aux_cp ? h : 0, aux_len, aux_enc);
	    if (aux_cp) {
		*cp++ = h;

		if (aux_pr) {
		    aux_pr->next = aux_cp->next;
		    aux_cp->next = aux_hist_p;
		    aux_hist_p = aux_cp;
		}
	    } else {
		*cp++ = 0;
		memcpy(cp, aux_enc, aux_len);
		cp += aux_len;
		*cp++ = 0;

		/* Take off free list */
		if (aux_hist_f) {
		    aux_hist_t *tmp = aux_hist_f;

		    aux_hist_f = tmp->next;
		    tmp->next = aux_hist_p;
		    aux_hist_p = tmp;
		} else {
		    aux_pr->next = aux_hist_p;
		    aux_hist_p = aux_pr;
		    aux_pr2->next = NULL;
		}

		aux_hist_p->len = aux_len;
		memcpy(aux_hist_p->aux, aux_enc, aux_len);
	    }

	    samp->delayed = -1;
	    for (j = 0; j < samp->a_cnt; j++) {
		if (samp->a[j].type == AUX_TYPE_RG)
		    continue;

#if 1
		if (samp->a[j].type == AUX_TYPE_OQ) {
		    samp->delayed = j;
		    continue;
		}
#endif

		switch (samp->a[j].format) {
		case 'i': {
		    int v = atoi(samp->a[j].val);
		    //printf("v=%d\n", v);
		    if (v > 0) {
			cp += int2u7(v, (unsigned char *)cp);
		    } else if (v < 0) {
			cp += int2u7(-v, (unsigned char *)cp);
		    }
		    break;
		}

		case 'A':
		    *cp++ = *samp->a[j].val;
		    break;

		case 'Z':
		    memcpy(cp, samp->a[j].val, samp->a[j].len);
		    cp += samp->a[j].len;
		    *cp++ = 0;
		    break;

		default:
		    fprintf(stderr, "Unknown aux format %c\n",
			    samp->a[j].format);
		    break;
		}
	    }
	}

#if 1
	/* Add delayed OQ values */
	for (i = 0; i < count; i++) {
	    //sam_entry *samp = &sam[i];
	    sam_entry *samp = ord[i].sam;

	    if ((j = samp->delayed) == -1)
		continue;

	    /* OQ has :Z: format */
	    memcpy(cp, samp->a[j].val, samp->a[j].len);
	    cp += samp->a[j].len;
	    *cp++ = 0;
	}
#endif

	/* List combinations of aux type & formats first */
	
	/* Then list entries by the combination index and encoded values */
#endif
	//write(2, cp2, cp-cp2);
	blk_size[13] = cp-cp2; cp2 = cp;
   }

	osz = cp-data;
	switch (mode) {
	case COMP_MODE_LZMA:
	    gzout = mem_deflate(COMP_MODE_LZMA, level, data, osz, &ssz);
	    //gzout = mem_deflate_parts(COMP_MODE_LZMA, level, data, blk_size, 14, &ssz);
	    free(data);
	    data = gzout;
	    break;

	case COMP_MODE_ZLIB:
	    gzout = mem_deflate_parts(COMP_MODE_ZLIB, level, data,
				      blk_size, 14, &ssz);
	    //gzout = mem_deflate(COMP_MODE_ZLIB, level, data, osz, &ssz);
	    free(data);
	    data = gzout;
	    break;

	default:
	    ssz = osz;
	}
	fprintf(stderr, "Compressed %d to %ld bytes\t%f\n",
		osz,  ssz, 100.0*ssz/osz);

	u32 = count; fwrite(&u32, 4, 1, fpo);
	u32 = ssz;   fwrite(&u32, 4, 1, fpo);
	fwrite(data, 1, ssz, fpo);
	//u32 = count; write(1, &u32, 4);
	//u32 = ssz;   write(1, &u32, 4);
	//write(1, data, ssz);

	free(data);

	for (i = 0; i <= nl; i++) {
	    free(lines[i]);
	}
    }

    /* Free it */
    free(sam);
    free(ord);
    free(pos);

    for (i = 0; i < count; i++)
	free(lines[i]);
    free(lines);
}

void usage(void) {
    fprintf(stderr, "Usage: sam_comp [options] < in.sam > out.zam\n");
    fprintf(stderr, "    -b nl         Process 'nl' lines per block.\n");
    fprintf(stderr, "    -c method     Compression method: zlib, lzma, none.\n");
    fprintf(stderr, "    -1 to -9      Compression level.\n");
    exit(1);
}

#define IO_BUF_SIZE 2*1024*1024

int main(int argc, char **argv) {
    int block_size = 10000;
    int mode = COMP_MODE_ZLIB;
    int level = 6;
    FILE *in = stdin, *out = stdout;
    char *in_buf  = malloc(IO_BUF_SIZE);
    char *out_buf = malloc(IO_BUF_SIZE);

    while (argc > 1) {
	if (strcmp(argv[1], "-c") == 0) {
	    if (argc < 2)
		usage();

	    if (strcmp(argv[2], "zlib") == 0) {
		mode = COMP_MODE_ZLIB;
	    } else if (strcmp(argv[2], "lzma") == 0) {
		mode = COMP_MODE_LZMA;
	    } else if (strcmp(argv[2], "none") == 0) {
		mode = COMP_MODE_NONE;
	    } else {
		usage();
	    }

	    argc--;
	    argv++;
	}

	else if (strcmp(argv[1], "-b") == 0) {
	    if (argc < 2)
		usage();

	    block_size = atoi(argv[2]);
	    argc--;
	    argv++;
	}
	    
	else if (argv[1][0] == '-' &&
		 argv[1][1] >= '1' && argv[1][1] <= '9') {
	    level = argv[1][1] - '0';
	}

	else if (argv[1][0] == '-')
	    usage();

	else
	    break;

	argc--;
	argv++;
    }

    if (argc >= 2) {
	if (NULL == (in = fopen(argv[1], "rb"))) {
	    perror(argv[1]);
	    exit(1);
	}
    }
    if (argc >= 3) {
	if (NULL == (out = fopen(argv[2], "wb"))) {
	    perror(argv[2]);
	    exit(1);
	}
    }

    setvbuf(in,  in_buf,  _IOFBF, IO_BUF_SIZE);
    setvbuf(out, out_buf, _IOFBF, IO_BUF_SIZE);

    comp(in, out, block_size, mode, level);

    if (in != stdin)
	fclose(in);
    if (out != stdout)
	fclose(out);

    if (1) {
	int i, tot = 0;
	for (i = 0; i < 14; i++)
	    tot += part_comp[i];

	for (i = 0; i < 14; i++)
	    fprintf(stderr, "Part %2d: %8d => %8d, perc %5.1f\n",
		    i, part_uncomp[i], part_comp[i],
		    100.0 * part_comp[i] / tot);
    }

    fflush(stdout);

    free(in_buf);
    free(out_buf);

    return 0;
}

/*

zlib compression parts on 10k lines/block:
1m.sam (54195436)
Part  0 => comp.size   420638, perc   0.8
Part  1 => comp.size   428376, perc   0.8
Part  2 => comp.size   243204, perc   0.4
Part  3 => comp.size   166920, perc   0.3
Part  4 => comp.size  1200213, perc   2.2
Part  5 => comp.size  1226247, perc   2.3
Part  6 => comp.size  4215493, perc   7.8
Part  7 => comp.size     3443, perc   0.0
Part  8 => comp.size   368146, perc   0.7
Part  9 => comp.size    46443, perc   0.1
Part 10 => comp.size    11573, perc   0.0
Part 11 => comp.size  4500269, perc   8.3
Part 12 => comp.size 39036628, perc  72.0 (4.0 bits/base)
Part 13 => comp.size  2327843, perc   4.3

100k.sam (2876689 bytes)
Part  0 => comp.size   149595, perc   5.2
Part  1 => comp.size    27752, perc   1.0
Part  2 => comp.size    40553, perc   1.4
Part  3 => comp.size    45939, perc   1.6
Part  4 => comp.size      332, perc   0.0
Part  5 => comp.size   128238, perc   4.5
Part  6 => comp.size   303840, perc  10.6
Part  7 => comp.size      338, perc   0.0
Part  8 => comp.size    12518, perc   0.4
Part  9 => comp.size      428, perc   0.0
Part 10 => comp.size      444, perc   0.0
Part 11 => comp.size   466311, perc  16.2
Part 12 => comp.size  1470804, perc  51.2 (3.1 bits/base)
Part 13 => comp.size   227830, perc   7.9

lzma versions:
1m.sam (50434520)
Part  0 => comp.size   355444, perc   0.7
Part  1 => comp.size   401568, perc   0.8
Part  2 => comp.size   240392, perc   0.5
Part  3 => comp.size   162756, perc   0.3
Part  4 => comp.size  1084208, perc   2.1
Part  5 => comp.size  1048568, perc   2.1
Part  6 => comp.size  3713216, perc   7.4
Part  7 => comp.size    11232, perc   0.0
Part  8 => comp.size   351352, perc   0.7
Part  9 => comp.size    53232, perc   0.1
Part 10 => comp.size    18244, perc   0.0
Part 11 => comp.size  3392488, perc   6.7
Part 12 => comp.size 37712288, perc  74.8
Part 13 => comp.size  1888660, perc   3.7

100k.sam (2471880 bytes, vs 2465000 encoded without ..._parts version)
Part  0 => comp.size   122012, perc   4.9
Part  1 => comp.size    27076, perc   1.1
Part  2 => comp.size    40288, perc   1.6
Part  3 => comp.size    46576, perc   1.9
Part  4 => comp.size     1080, perc   0.0
Part  5 => comp.size   106392, perc   4.3
Part  6 => comp.size   227996, perc   9.2
Part  7 => comp.size     1120, perc   0.0
Part  8 => comp.size    13388, perc   0.5
Part  9 => comp.size     1240, perc   0.1
Part 10 => comp.size     1160, perc   0.0
Part 11 => comp.size   366824, perc  14.8
Part 12 => comp.size  1315348, perc  53.2
Part 13 => comp.size   200192, perc   8.1

=> quality is key to compress.

comp1 (1st order range coding) seems good compression ratio on large data
set, but not on small:

comp1 1m.q   => 35040092
comp2 1m.q   => 34285760
xz    1m.q   => 37551844 
gzip  1m.q   => 40987188 (why more than above? mode/break points I guess)
lpaq  1m.q   => 33408659

comp1 100k.q => 1686842
comp2 100k.q => 1656128
xz    100k.q => 1298864
gzip  100k.q => 1530459
lpaq  100k.q => 1292333

Ie order-2 range coding works very well on quality data unless
redundancy can be found, whereupon lz based systems win out.
Combination of both to win?

 */
